#include "pathmodel.h"
#define TIME 100 // time for path: 50 seconds
#define AXIS 1 // 320 mm axis
#define GRADMIN 0.00000000001
#define STARTGRAD 0.01
#define NVARS 12
#define AMP 0.3
#define MAXITER 1000

typedef struct
{
	double N;
	double StartAngle;
	double *XM;
	double *YM;
	SpeedProfile P;
	RcRcModel AgoR;
	RcRcModel AgoL;
	RcRcModel AntR;
	RcRcModel AntL;
}Target;

PathModel::PathModel(double *X, double *Y, double Angle, int NP)
{
	int i;
	
	this->N = (double)(NP); // set number of points on path
	this->Path.dT = (double)(TIME) / N;

	this->StartAngle = Angle;

	// allocate memory for Path
	Path.Vr 	= new double[this->N];
	Path.Vl 	= new double[this->N];
	Path.Angle 	= new double[this->N];
	Path.X 		= new double[this->N];
	Path.Y 		= new double[this->N];
//	Path.dT		= (ModelRight.UB-ModelRight.LB)/this->N;

	XMJ = new double[N];
	YMJ = new double[N];

	for(i = 0;i < N;i++)
	{
		XMJ[i] = X[i];
		YMJ[i] = Y[i];
	}

	// starting values of the models
	AgoR.T1  = 0.17;
	AgoL.T1	 = 0.157;
	AgoR.T2  = 0.1603;
	AgoL.T2	 = 0.15701;
	AgoL.Amp = AMP;
	AgoR.LB  = 0;
	AgoL.LB	 = 0;
	AgoR.UB  = 5;
	AgoL.UB	 = 5;
	AgoR.Amp = AMP;	

	AntR.T1  = 0.18;
	AntL.T1	 = 0.167;
	AntR.T2  = 0.1703;
	AntL.T2	 = 0.16701;
	AntL.Amp = AMP;
	AntR.LB  = 0;
	AntL.LB	 = 0;
	AntR.UB  = 5;
	AntL.UB	 = 5;
	AntR.Amp = AMP;	
	Fill();
	Solve();
}

void PathModel::CalcVec(double Vr, double Vl, double Angle, double X, double Y, double *AngleNew, double *XNew, double *YNew)
{
	double T = this->Path.dT;
	
	if(Vr == Vl)
	{ // speeds equal, so Dx/y = v*t*direction, t is Dt
		*XNew = X + T*Vr*cos(Angle);
		*YNew = Y + T*Vl*sin(Angle);
		*AngleNew = Angle;
	} else
	{	
		double
			C = (AXIS*(Vr+Vl))/(2*(Vr-Vl));
			
		*AngleNew = Angle + ((Vr - Vl)/AXIS);
		*XNew = X + C * (sin(*AngleNew)*T - sin(Angle)); // here we need the new and old Phi
		*YNew = Y - C * (cos(*AngleNew)*T - cos(Angle)); // hence the introduction of Phi_new	
	}
}

double PathModel::RcRc(double T1, double T2, double Amp, double T3, double T4, double Amp1, double t)
{
	double a;
	double b;
	double c;
	double d;
	 
	a=(T1-T2)*(T1-T3)*(T1-T4);
	b=(T1-T2)*(T2-T3)*(T2-T4);
	c=(T1-T3)*(T3-T2)*(T3-T4);
	d=(T1-T4)*(T4-T2)*(T4-T3);

	return(Amp * Amp * Amp1 * Amp1 *
				( 
					-(exp(-t*T1)/a)
					+(exp(-t*T2)/b)
					+(exp(-t*T3)/c)
					+(exp(-t*T4)/d)
				)
			);
}

void PathModel::Fill()
{
	for(int i = 0;i < N;i++) // for loop to fill in results 
	{
		double t = (this->Path.dT*(double)(i));
		
		Path.Vr[i] = RcRc(AgoR.T1, AgoR.T2, AgoR.Amp, AntR.T1, AntR.T2, AntR.Amp, t);
		Path.Vl[i] = RcRc(AgoL.T1, AgoL.T2, AgoL.Amp, AntL.T1, AntL.T2, AntL.Amp, t);
		Path.Angle[i] = this->StartAngle;
 
		//cout << Path.Vl[i] << "\t" << Path.Vr[i] << endl;

		if(i > 0) 	CalcVec(Path.Vr[i], Path.Vl[i], Path.Angle[i-1], Path.X[i-1], Path.Y[i-1], &Path.Angle[i], &Path.X[i], &Path.Y[i]); 
		else 		CalcVec(Path.Vr[i], Path.Vl[i], this->StartAngle,0,0, &Path.Angle[i], &Path.X[i], &Path.Y[i]); 
		
	}
}


RcRcModel PathModel::GetAgoR()
{ // returns current RcRcmodel in member "Model"
	RcRcModel result;
	result = this->AgoR;
	return(result); // can one return this->Model immediately ?
}

RcRcModel PathModel::GetAgoL()
{ // returns current RcRcmodel in member "Model"
	RcRcModel result;
	result = this->AgoL;
	return(result); // can one return this->Model immediatly ?
}

RcRcModel PathModel::GetAntR()
{ // returns current RcRcmodel in member "Model"
	RcRcModel result;
	result = this->AntR;
	return(result); // can one return this->Model immediately ?
}

RcRcModel PathModel::GetAntL()
{ // returns current RcRcmodel in member "Model"
	RcRcModel result;
	result = this->AntL;
	return(result); // can one return this->Model immediatly ?
}

SpeedProfile PathModel::GetPath()
{ // returns current SpeedProfile in member "Path"
	int i;
	SpeedProfile result;

	result.Vr		= new double[N];
	result.Vl		= new double[N];
	result.Angle	= new double[N];
	result.X		= new double[N];
	result.Y		= new double[N];

	for(i = 0;i < N;i++)
	{
		result.Vr[i]	= Path.Vr[i];
		result.Vl[i]	= Path.Vl[i];
		result.Angle[i] = Path.Angle[i];
		result.X[i]		= Path.X[i];
		result.Y[i]		= Path.Y[i];
		result.dT		= Path.dT;
	}
	return(result);
}


void CalcVecSolve(double Vr, double Vl, double Angle, double X, double Y, double *AngleNew, double *XNew, double *YNew, double T)
{	
	if(Vr == Vl)
	{ // speeds equal, so Dx/y = v*t*direction, t is Dt
		*XNew = X + T*Vr*cos(Angle);
		*YNew = Y + T*Vl*sin(Angle);
		*AngleNew = Angle;
	} else
	{	
		double
			C = (AXIS*(Vr+Vl))/(2*(Vr-Vl));
			
		*AngleNew = Angle + ((Vr - Vl)/AXIS);
		*XNew = X + C * (sin(*AngleNew)*T - sin(Angle)); // here we need the new and old Phi
		*YNew = Y - C * (cos(*AngleNew)*T - cos(Angle)); // hence the introduction of Phi_new	
	}
}

double RcRcSolve(double T1, double T2, double Amp, double T3, double T4, double Amp1, double t)
{
	double a;
	double b;
	double c;
	double d;


	a=(T1-T2)*(T1-T3)*(T1-T4);
	b=(T1-T2)*(T2-T3)*(T2-T4);
	c=(T1-T3)*(T3-T2)*(T3-T4);
	d=(T1-T4)*(T4-T2)*(T4-T3);

//	cout << (T1-T2) << "|" << T1 << "-" << T3 << "=" << (T1-T3) << "|" << (T1-T4) << "|" << endl;
//	cout << (T1-T2) << "|" << (T2-T3) << "|" << (T2-T4) << "|" << endl;
//	cout << (T1-T3) << "|" << (T3-T2) << "|" << (T3-T4) << "|" << endl;
//	cout << (T1-T4) << "|" << (T4-T2) << "|" << (T4-T3) << "|" << endl;
//	cout << a << "|" << b << "|" << c << "|" << d << "|" << endl;
//	cout << T1 << "|" << T2 << "|" << T3 << "|" << T4 << "|" << endl;

	return(Amp * Amp * Amp1 * Amp1 *
				( 
					-(exp(-t*T1)/a)
					+(exp(-t*T2)/b)
					+(exp(-t*T3)/c)
					+(exp(-t*T4)/d)
				)
			);
}

void FillSolve(Target *T)
{
	for(int i = 0;i < T->N;i++) // for loop to fill in results 
	{
		double t = (T->P.dT*(double)(i));
		

		T->P.Vr[i] = RcRcSolve(T->AgoR.T1, T->AgoR.T2, T->AgoR.Amp, T->AntR.T1, T->AntR.T2, T->AntR.Amp, t);
		T->P.Vl[i] = RcRcSolve(T->AgoL.T1, T->AgoL.T2, T->AgoL.Amp, T->AntL.T1, T->AntL.T2, T->AntL.Amp, t);
		T->P.Angle[i] = T->StartAngle;

 
		if(i > 0) 	CalcVecSolve(T->P.Vr[i], T->P.Vl[i], T->P.Angle[i-1], T->P.X[i-1], T->P.Y[i-1], &T->P.Angle[i], &T->P.X[i], &T->P.Y[i], T->P.dT); 
		else 		CalcVecSolve(T->P.Vr[i], T->P.Vl[i], T->StartAngle,0,0, &T->P.Angle[i], &T->P.X[i], &T->P.Y[i],T->P.dT); 
	}
}


double F(const gsl_vector *B, void *Tar)
{
    Target *T = (Target *)Tar;
	double result = 0;

	T->AgoR.T1  = gsl_vector_get(B, 0);
	T->AgoR.T2  = gsl_vector_get(B, 1);
	T->AgoR.Amp = gsl_vector_get(B, 2);
	T->AgoL.T1  = gsl_vector_get(B, 3);
	T->AgoL.T2  = gsl_vector_get(B, 4);
	T->AgoL.Amp = gsl_vector_get(B, 5);

	T->AntR.T1  = gsl_vector_get(B, 6);
	T->AntR.T2  = gsl_vector_get(B, 7);
	T->AntR.Amp = gsl_vector_get(B, 8);
	T->AntL.T1  = gsl_vector_get(B, 9);
	T->AntL.T2  = gsl_vector_get(B, 10);
	T->AntL.Amp = gsl_vector_get(B, 11);

	FillSolve(T);
			
	for(int i = 1;i < (T->N-1);i++)
	{
		result +=1;
//		cout << result << "|" << T->P.X[i] << "|" << T->XM[i] << "|" << T->XM[i] << "|" << T->P.X[i] << endl;
		result+=((T->P.X[i] - T->XM[i])*(T->P.X[i] - T->XM[i])+0.01); // what is best ?
		result+=((T->P.Y[i] - T->YM[i])*(T->P.Y[i] - T->YM[i])+0.01);
//		result+=(T->XM[i] - T->P.X[i])*(T->XM[i] - T->P.X[i]);
//		result+=(T->YM[i] - T->P.Y[i])*(T->YM[i] - T->P.Y[i]);
	}

	return result;
}

void FDF(const gsl_vector *Vec, void *Param, gsl_vector *dVec)
{
	double *P = (double *)Param;

	gsl_multimin_function FMF ;
	FMF.f = F;
	FMF.params = P;
	FMF.n = Vec->size;
	gsl_multimin_diff (&FMF, Vec, dVec);

}

void FFDF (const gsl_vector *x, void *params, double *f, gsl_vector *df) 
{
  *f =F(x, params); 
  FDF(x, params, df);
}



void PathModel::Solve()
{ // Solve model using bfgs algorithm
	size_t iter = 0;
    int status = 0;
	int i = 0;

	Target *P = new Target;

	P->N			= N;
	P->StartAngle	= this->StartAngle;

	P->XM		= new double[P->N];
	P->YM		= new double[P->N];

	P->P.Vr		= new double[P->N];
	P->P.Vl		= new double[P->N];
	P->P.Angle	= new double[P->N];
	P->P.X		= new double[P->N];
	P->P.Y		= new double[P->N];
	P->P.dT		= this->Path.dT;

	P->AgoR = AgoR; 
	P->AgoL = AgoL;
	P->AntR = AntR; 
	P->AntL = AntL;

	for(i = 0;i < P->N;i++)
	{
		P->XM[i]		= this->XMJ[i];
		P->YM[i]		= this->YMJ[i];
		P->P.Vr[i]		= this->Path.Vr[i];
		P->P.Vl[i]		= this->Path.Vl[i];
		P->P.Angle[i]	= this->Path.Angle[i];
		P->P.X[i]		= this->Path.X[i];
		P->P.Y[i]		= this->Path.Y[i];
	}

	const gsl_multimin_fdfminimizer_type *T;
	gsl_multimin_fdfminimizer *s;
    gsl_multimin_function_fdf mf;
    gsl_vector *Start;

    mf.f = &F; // dunno how to point to member that is not static ... 
    mf.df = &FDF;
    mf.fdf = &FFDF;
    mf.n = NVARS;
    mf.params = (void*)P;

    Start = gsl_vector_alloc(NVARS);

	gsl_vector_set(Start, 0, P->AgoR.T1);
	gsl_vector_set(Start, 1, P->AgoR.T2);
	gsl_vector_set(Start, 2, P->AgoR.Amp);
	gsl_vector_set(Start, 3, P->AgoL.T1);
	gsl_vector_set(Start, 4, P->AgoL.T2);
	gsl_vector_set(Start, 5, P->AgoL.Amp);

	gsl_vector_set(Start, 6,  P->AntR.T1);
	gsl_vector_set(Start, 7,  P->AntR.T2);
	gsl_vector_set(Start, 8,  P->AntR.Amp);
	gsl_vector_set(Start, 9,  P->AntL.T1);
	gsl_vector_set(Start, 10, P->AntL.T2);
	gsl_vector_set(Start, 11, P->AntL.Amp);


    T = gsl_multimin_fdfminimizer_vector_bfgs;
    s = gsl_multimin_fdfminimizer_alloc (T, NVARS);

    gsl_multimin_fdfminimizer_set(s, &mf, Start, STARTGRAD, 1e-4);

	do
      {
		if(iter==1) cout << s->f << " *********** Start cost function for muscle model fit" <<endl;
        iter++;
        status = gsl_multimin_fdfminimizer_iterate(s);				

	    if (status) 
		{
			if (status == GSL_EINVAL) 
			{
				fprintf (stderr, "invalid argument, iter=%d\n", iter);
			} else 
			{
				fprintf (stderr, "failed, gsl_errno=%d\n", status);
			}
			exit (-1);
		}

/*
        if (status)
		{
			cout << "Error causing status: "  << status;
            break;
		}*/

       status = gsl_multimin_test_gradient(s->gradient, GRADMIN);

        if (status == GSL_SUCCESS)
		{
            cout <<"#Minimum found for muscle model:";
		} else 
		{
		//	cout << iter << ". ";
			for(i = 0;i < NVARS;i++)
			{		
			//	printf ("%f ",gsl_vector_get(s->x,i) );
		//		cout << gsl_vector_get(s->x,i) << "\t";
			}// cout << " g: " << s->gradient <<endl;
		}

      }
    while (status == GSL_CONTINUE && iter < MAXITER);

	cout << s->f <<" *********** End cost function for muscle model fit, no minimum found" << endl;

	this->AgoR.T1  =gsl_vector_get(s->x,0);
	this->AgoR.T2  =gsl_vector_get(s->x,1);
	this->AgoR.Amp =gsl_vector_get(s->x,2);
	this->AgoL.T1  =gsl_vector_get(s->x,3);
	this->AgoL.T2  =gsl_vector_get(s->x,4);
	this->AgoL.Amp =gsl_vector_get(s->x,5);

	this->AntR.T1  =gsl_vector_get(s->x,6);
	this->AntR.T2  =gsl_vector_get(s->x,7);
	this->AntR.Amp =gsl_vector_get(s->x,8);
	this->AntL.T1  =gsl_vector_get(s->x,9);
	this->AntL.T2  =gsl_vector_get(s->x,10);
	this->AntL.Amp =gsl_vector_get(s->x,11);

	this->Fill();

    gsl_multimin_fdfminimizer_free(s);
    gsl_vector_free(Start);
}